@@ -11,15 +11,17 @@ import pytest
 from bdd_coder import exceptions
 from bdd_coder.features import StepSpec
 from bdd_coder import stock
-from bdd_coder.text_utils import OK, FAIL, PENDING, TO, COMPLETION_MSG, BOLD, Style, indent
+from bdd_coder.text_utils import (
+    OK, FAIL, PENDING, TO, COMPLETION_MSG, BOLD, Style, indent, ExcInfo)
 
 
 class StepRun(stock.Repr):
-    def __init__(self, scenario_run, step, kwargs):
+    def __init__(self, step, scenario_run):
         self.scenario_run = scenario_run
         self.step = step
-        self.kwargs = kwargs
+        self.kwargs = {}
         self.result = None
+        self.is_last = False
 
     @property
     def symbol(self):
@@ -29,11 +31,10 @@ class StepRun(stock.Repr):
     def symbol(self, value):
         self.__symbol = value
         self.__end_time = datetime.datetime.utcnow()
+        self.log()
 
-        if value == FAIL:
-            self.scenario_run.symbol = FAIL
-        elif value == OK and self.step.is_last:
-            self.scenario_run.symbol = OK
+        if value == FAIL or value == OK and self.is_last:
+            self.scenario_run.symbol = value
 
     @property
     def end_time(self):
@@ -49,14 +50,13 @@ class StepRun(stock.Repr):
 
     @property
     def formatted_result(self):
-        if isinstance(self.result, tuple) and self.result:
-            if self.symbol == OK:
-                text = '\n'.join([f'    {repr(v)}' for v in self.result])
+        if isinstance(self.result, tuple) and self.result and self.symbol == OK:
+            text = '\n'.join([f'    {repr(v)}' for v in self.result])
 
-                return f'\n  {TO} {text.lstrip()}'
+            return f'\n  {TO} {text.lstrip()}'
 
-            if self.symbol == FAIL:
-                return f'{TO} {self.result[2]}'
+        if self.symbol == FAIL:
+            return f'{TO} {self.result.highlighted_traceback}'
 
         return ''
 
@@ -68,21 +68,27 @@ class StepRun(stock.Repr):
 
 
 class ScenarioRun(stock.Repr):
-    def __init__(self, test_run, scenario):
-        super().__init__()
-        self.test_run = test_run
+    def __init__(self, test_id, scenario, parent_run=None):
+        self.test_id = test_id
         self.scenario = scenario
-        self.runs = []
+        self.parent_run = parent_run
+        self.is_last = False
+        self.runs = [StepRun(step, self) if step.doc_scenario is None else ScenarioRun(
+            test_id, step.doc_scenario, self) for step in scenario.steps]
+        self.runs[-1].is_last = True
 
     def __iter__(self):
-        yield from self.runs
+        yield self
+
+        for run in self.runs:
+            if isinstance(run, ScenarioRun):
+                yield from run
 
     def __str__(self):
         qualname = self.scenario.qualname
 
         if self.symbol == FAIL:
-            exc_type, exc_value, tb_text = self.result
-            result_text = f' {TO} {exc_type.__name__}: {exc_value}'
+            result_text = f' {TO} {self.result.exc_type.__name__}: {self.result.exc_value}'
         elif self.symbol == OK:
             result_text = f' {TO} {self.result}' if self.result else '.'
 
@@ -91,7 +97,10 @@ class ScenarioRun(stock.Repr):
 
     @property
     def result(self):
-        return self.runs[-1].result if self.symbol != PENDING else ''
+        for step_run in self.iter_step_runs():
+            if step_run.symbol == FAIL:
+                return step_run.result
+        return step_run.result
 
     @property
     def symbol(self):
@@ -101,9 +110,10 @@ class ScenarioRun(stock.Repr):
     def symbol(self, value):
         self.__symbol = value
         self.__end_time = datetime.datetime.utcnow()
+        self.log()
 
-        if value == FAIL and not self.scenario.is_test:
-            self.test_run.symbol = FAIL
+        if self.parent_run is not None and (value == FAIL or value == OK and self.is_last):
+            self.parent_run.symbol = value
 
     @property
     def end_time(self):
@@ -113,53 +123,22 @@ class ScenarioRun(stock.Repr):
     def end_time(self, value):
         raise AttributeError("'end_time' is read-only")
 
-    def append_run(self, step, kwargs):
-        self.runs.append(StepRun(self, step, kwargs))
+    def iter_step_runs(self):
+        for run in self.runs:
+            if isinstance(run, StepRun):
+                yield run
+            elif isinstance(run, ScenarioRun):
+                yield from run.iter_step_runs()
+
+    def get_pending_step_run(self, step):
+        for step_run in self.iter_step_runs():
+            if step_run.step == step and step_run.symbol == PENDING:
+                return step_run
 
     def log(self):
-        self.scenario.gherkin.logger.info(f'└─{self}')
-
-
-class TestRun(stock.Repr):
-    def __init__(self, test_id, scenario):
-        self.test_id = test_id
-        self.scenario = scenario
-        self.runs = collections.defaultdict(list)
-        self.add_run(self.scenario)
-
-    def __str__(self):
-        return (f'{PENDING} ' if self.symbol == PENDING else
-                f'{self.end_time} {BOLD[self.symbol]} ') + self.test_id
-
-    def __iter__(self):
-        yield from self.sorted_runs
-
-    @property
-    def sorted_runs(self):
-        return sorted(itertools.chain(*self.runs.values()), key=lambda sr: sr.end_time)
-
-    @property
-    def scenario_run(self):
-        return self.runs[self.scenario.name][-1]
-
-    @property
-    def symbol(self):
-        return self.scenario_run.symbol
-
-    @symbol.setter
-    def symbol(self, value):
-        self.scenario_run.symbol = value
-
-    @property
-    def end_time(self):
-        return self.scenario_run.end_time
-
-    @property
-    def result(self):
-        return self.scenario_run.result
-
-    def add_run(self, scenario):
-        self.runs[scenario.name].append(ScenarioRun(self, scenario))
+        self.scenario.gherkin.logger.info('└─' + (
+            f'{PENDING} {self.scenario.qualname}' if self.symbol == PENDING else
+            f'{self.end_time} {BOLD[self.symbol]} {self.scenario.qualname}'))
 
 
 class Step(StepSpec):
@@ -168,8 +147,6 @@ class Step(StepSpec):
         self.scenario = scenario
         self.doc_scenario = None
         self.test_scenario = None
-        self.is_last = False
-        self.is_first = False
         self.method_qualname = ''
 
     @property
@@ -195,17 +172,14 @@ class Step(StepSpec):
             if tester.current_run.symbol != PENDING:
                 return
 
-            if self.is_first and not self.scenario.is_test:
-                tester.current_run.add_run(self.scenario)
-
-            scenario_run = tester.current_run.runs[self.scenario.name][-1]
-            scenario_run.append_run(self, kwargs)
-            step_run = scenario_run.runs[-1]
+            step_run = tester.current_run.get_pending_step_run(self)
+            step_run.kwargs = {k: v for k, v in kwargs.items()
+                               if k not in self.gherkin.fixtures_not_to_log}
 
             try:
                 step_run.result = step_method(tester, *args, **kwargs)
             except Exception:
-                step_run.result = exceptions.format_next_traceback()
+                step_run.result = ExcInfo()
                 step_run.symbol = FAIL
             else:
                 step_run.symbol = OK
@@ -214,11 +188,6 @@ class Step(StepSpec):
                     for name, value in zip(self.output_names, step_run.result):
                         self.gherkin.outputs[name].append(value)
 
-            step_run.log()
-
-            if scenario_run.symbol != PENDING:
-                scenario_run.log()
-
         return pytest.fixture(name=self.fixture_name, params=self.fixture_param)(
             logger_step_method)
 
@@ -269,14 +238,6 @@ class Scenario(stock.Repr):
                 name=self.name, positions=', '.join([str(i) for i in wrong_values]),
                 length=len(param_ids))
 
-        simple_steps = list(filter(lambda s: s.doc_scenario is None, self.steps))
-
-        if simple_steps:
-            simple_steps[0].is_first = True
-
-            if simple_steps[-1] == self.steps[-1]:
-                self.steps[-1].is_last = True
-
         for step in self.steps:
             if step.doc_scenario is None:
                 fine_steps.append(step)
@@ -318,7 +279,7 @@ class Scenario(stock.Repr):
             __tracebackhide__ = True
 
             if tester.current_run.symbol == FAIL:
-                pytest.fail(msg=tester.current_run.result[2], pytrace=False)
+                pytest.fail(msg=tester.current_run.result.next_traceback, pytrace=False)
 
         if len(param_ids) == 1:
             param_values = [v[0] for v in param_values]
@@ -343,12 +304,14 @@ class Scenario(stock.Repr):
 
 
 class Gherkin(stock.Repr):
-    def __init__(self, aliases, validate=True, **logging_kwds):
+    def __init__(self, aliases, validate=True, fixtures_not_to_log=('request',),
+                 **logging_kwds):
         self.reset_logger(**logging_kwds)
         self.reset_outputs()
         self.scenarios = collections.defaultdict(dict)
         self.aliases = aliases
         self.validate = validate
+        self.fixtures_not_to_log = fixtures_not_to_log
         self.test_runs = {}
 
     def __str__(self):
@@ -379,7 +342,7 @@ class Gherkin(stock.Repr):
             yield from self.scenarios[class_name].values()
 
     def new_run(self, test_id, scenario):
-        self.test_runs[test_id] = TestRun(test_id, scenario)
+        self.test_runs[test_id] = ScenarioRun(test_id, scenario)
         self.logger.info('_'*26)
 
     def reset_logger(self, logs_path, maxBytes=100000, backupCount=10):
